#!/usr/bin/env python

# ZAVRSNI RAD br. 1988
# Automatska paralelizacija uporabom genetskog programiranja
# Ivo Majic
# v1.0

import itertools
import random
import bisect
import copy
import re

# Klasa 'node' u sebi sadrzava informaciju o sadrzaju jednog cvora (ili lista)
# stabla koje predstavlja kromosom jedinke koja je opisana klasom 'individual'.

class node:
    
    def __init__(self):
        self.left = None
        self.right = None
        self.dubina = None
        self.data = None
    
    def grow(self, maksdubina, dubina, metoda):
        terminali = ('NULL', 'PARNULL')
        unarfunkcije = ('FPAR', 'FSEQ', 'LPAR', 'LSEQ', 'SHIFT')
        self.dubina = dubina
        
        if dubina == 0:
            self.data = self.function()
        
        elif dubina >= maksdubina:
            self.data = random.choice(terminali)
            return
        
        elif metoda == 'full':
            self.data = self.function()
        
        elif metoda == 'grow':
            if random.choice(('funkcija', 'terminal')) == 'terminal':
                self.data = random.choice(terminali)
                return
            else:
                self.data = self.function()
                   
        if self.data not in unarfunkcije:
            self.right = node()
            self.right.grow(maksdubina, dubina+1, metoda)
        self.left = node()
        self.left.grow(maksdubina, dubina+1, metoda)
            
    def function(self):
        funkcije = ('FPAR', 'FSEQ', 'P', 'S', 'LPAR', 'LSEQ', 'SHIFT')
        izbor = random.choice(funkcije)
        if izbor in ('P','S'):
            # Inicijalne vrijednosti te vjerojatnosti PXX/SXX transformacija
            trans = (50, 25, 75, 33, 66, 'RAND')
            intervali = (25,40,55,70,85,100)
            data = trans[bisect.bisect_left(intervali, random.randint(0,100))]
            if data == 'RAND':
                return izbor + str(random.randint(1,99))
            else:
                return izbor + str(data)
        else:
            return izbor

# Klasa 'individual' u sebi sadrzava genetski zapis jedinke (kromosom), te
# ocjene dobrote jedinke po pitanju korektnosti i paralelnosti.

class individual:
    
    def __init__(self):
        self.root = node()
        self.korektnost = 0
        self.paralelnost = 0
        self.zapis = None
        
    def __cmp__(self, other):
        return cmp((self.korektnost, self.paralelnost),
                   (other.korektnost, other.paralelnost))
    
    # Metoda grow(maksdubina) 'ramped half-and-half' metodom bira koja ce metoda
    # izgradnje stabla biti koristena za trenutnu jedinku uz zadani parametar
    # 'maksdubina'.

    def grow(self, maksdubina):
        self.maksdubina = maksdubina
        metoda = random.choice(('full', 'grow'))
        maksdubina = random.choice(range(1, maksdubina+1))
        self.root.grow(maksdubina, 0, metoda)
    
    # Metoda 'mutate' bira nasumicni cvor u stablu, brise podstablo koje cvor
    # referencira te generira novo podstablo.
    
    def mutate(self, mprob):
        if random.random() <= mprob:
            nodes = self.nodelist()
            point = random.choice(nodes)
            metoda = random.choice(('full', 'grow'))
            point.grow(self.maksdubina, point.dubina, metoda)
            
    # Metoda 'refresh' osvjezava atribut 'dubina' koji sadrzi informaciju o
    # trenutnoj dubini na kojoj se nalazi cvor.
    
    def refresh(self, *args):
        try:
            node, dubina = args
        except:
            node, dubina = self.root, 0
            
        node.dubina = dubina
        if node.right != None:
            self.refresh(node.right, dubina+1)
        if node.left != None:
            self.refresh(node.left, dubina+1)
            
    # Metoda 'nodelist' vraca listu trenutnih cvorova u stablu.
    
    def nodelist(self, *args):
        try:
            node = args[0]
        except:
            node = self.root
            
        if node.data not in ('NULL', 'PARNULL'):
            unarfunkcije = ('FPAR', 'FSEQ', 'LPAR', 'LSEQ', 'SHIFT')
            if node.data in unarfunkcije:
                return self.nodelist(node.left)+[node]
            else:
                return self.nodelist(node.left)+[node]+self.nodelist(node.right)
        else:
            return []

# Klasa 'paragen' implementira steady-state genetski algoritam koji obavlja
# generiranje incijalne populacije jedinki (klasa 'individual')
# 'ramped half-and-half' metodom, evaluaciju dobrote jedinki te evoluciju
# populacije kroz vise iteracija.

class paragen:
    
    def __init__(self, btable):
        self.best = None
        self.btable = btable
    
    def parametri(self, args):
        self.iteracije, self.stagnacija, self.psize, self.cprob, self.mprob, self.maksdubina = args
    
    # Metoda 'fitness' kao parametar prima jedinku te joj odreduje dobrotu po
    # dva kriterija - korektnost i paralelnost
    
    def fitness(self, jedinka):
        jedinka.korektnost = 0
        jedinka.paralelnost = 0
        pseudotime = [[i, i] for i in xrange(len(self.btable))]
        self.transform(jedinka.root, pseudotime)
        
        values = set(map(lambda x:x[1], pseudotime))
        parallel = [[y[0] for y in pseudotime if y[1]==x] for x in values]
            
        # Provjeravamo jesu li sve instrukcije koje se izvode paralelno neovisne
        for group in parallel:
            if len(group) > 1:
                for instA, instB in itertools.combinations(group, 2):
                    jedinka.korektnost += self.bernstein(instA, instB)
                    
        # Provjeravamo jeli sacuvan poredak instrukcija koje su medusobno ovisne
        for index, (instA, timeA) in enumerate(pseudotime):
            for instB, timeB in pseudotime[index+1:]:
                if timeB < timeA:
                    jedinka.korektnost += self.bernstein(instA, instB)
                    
        jedinka.paralelnost = len(parallel)
        jedinka.zapis = parallel
        return
    
    # Metoda 'transform' provodi tranformacije zapisane u kromosomu (stablu) nad
    # slijednim instrukcijama.
    
    def transform(self, root, segment):
        if not segment: return
        
        if root.data == 'FSEQ':
            B = segment[1:]
            self.transform(root.left, B)
            return

        elif root.data == 'FPAR':
            A = segment[0]
            B = segment[1:]
            try:
                A[1] = B[0][1]
            except:
                pass
            self.transform(root.left, B)
            return

        elif root.data == 'LSEQ':
            B = segment[:-1]
            self.transform(root.left, B)
            return

        elif root.data == 'LPAR':
            B = segment[:-1]
            try:
                A = segment[-1]
                A[1] = B[0][1]
            except:
                pass
            self.transform(root.left, B)
            return

        elif root.data == 'SHIFT':
            B = segment
            for instB in B: instB[1] += 1
            self.transform(root.left, B)
            return

        elif root.data[0] == 'P' and root.data[1:].isdigit():
            split = int(round(float('0.'+root.data[1:])*len(segment)))
            A = segment[:split]
            B = segment[split:]
            if A:
                timeA = A[0][1]
                for instB in B:
                    instB[1] = timeA
                    timeA += 1
            self.transform(root.left, A)
            self.transform(root.right, B)
            return
        
        elif root.data[0] == 'S' and root.data[1:].isdigit():
            split = int(round(float('0.'+root.data[1:])*len(segment)))
            A = segment[:split]
            B = segment[split:]
            self.transform(root.left, A)
            self.transform(root.right, B)
            return

        elif root.data == 'NULL':
            B = segment[1:]
            timeA = segment[0][1]
            for instB in B:
                timeA += 1
                instB[1] = timeA
            return

        elif root.data == 'PARNULL':
            B = segment[1:]
            timeA = segment[0][1]
            for instB in B: instB[1] = timeA
            return

    # Metoda 'bernstein' nad dvije instrukcije provodi Bernsteinove uvjete kako
    # bi utvrdili mogu li se instrukcije izvrsavati paralelno.
    
    def bernstein(self, inst1, inst2):
        A = self.btable[inst1]
        B = self.btable[inst2]
        if not ((A[0]&B[1]) | (A[1]&B[0]) | (A[0]&B[0])): return 0
        return 1
    
    # Metoda 'selection' implementira genetski operator selekcije - 4-turnirska
    # selekcija.
    
    def selection(self, P):
        players = []
        for index in xrange(4):
            choice = random.randint(0, len(P)-1)
            players.append(P.pop(choice))
        
        players.sort(key=lambda x: (x.korektnost, x.paralelnost))
        
        mom = players[0]
        dad = players[1]
        
        return (mom, dad)
    
    # Metoda 'crossover' implementira genetski operator krizanja gdje se kod
    # dvije roditeljske jedinke zamjenjuju dva nasumicno odabrana podstabla.
    # Parametar 'cprob' odreduje hoce li se krizanje dogoditi.
    
    def crossover(self, jedinka1, jedinka2):
        mom = copy.deepcopy(jedinka1)
        dad = copy.deepcopy(jedinka2)
    
        momList = mom.nodelist()
        dadList = dad.nodelist()
        
        point1 = random.choice(momList)
        point2 = random.choice(dadList)
        
        child1 = mom
        child2 = dad
        
        unarfunkcije = ('FPAR', 'FSEQ', 'LPAR', 'LSEQ', 'SHIFT')
        choice1, choice2 = 'left', 'left'
        if point1.data not in unarfunkcije:
            choice1 = random.choice(('left','right'))
        if point2.data not in unarfunkcije:
            choice2 = random.choice(('left','right'))
            
        tmp1 = getattr(point1, choice1)
        tmp2 = getattr(point2, choice2)
        setattr(point1, choice1, tmp2)
        setattr(point2, choice2, tmp1)
        
        child1.refresh()
        child2.refresh()
        
        return (child1, child2)
    
    # Metoda 'run' pokrece sam algoritam te prati dolazi li do stagnacije
    # poboljsanja rezultata te prekida algoritam dosegene li stagnacija
    # odredenu vrijednost.
    
    def run(self):
        # Stvaranje inicijalne populacije
        P = [individual() for i in xrange(self.psize)]
        for jedinka in P:
            jedinka.grow(self.maksdubina)
            self.fitness(jedinka)
        # Pronalazenje inicijalno najbolje jedinke zbog pracenja stagnacije
        temp_best = P[0]
        for jedinka in P:
            if temp_best < jedinka: temp_best = jedinka
        briter, brstag = 0, 0
        while (briter < self.iteracije) and (brstag < self.stagnacija):
            briter +=1; brstag += 1
            if random.random() <= self.cprob:
                mom, dad = self.selection(P)
                child1, child2 = self.crossover(mom, dad)
                child1.mutate(self.mprob); child2.mutate(self.mprob)
                self.fitness(child1); self.fitness(child2)
                for jedinka in (mom, dad, child1, child2):
                    P.insert(random.randint(0, len(P)-1), jedinka)
                # Provjera dogada li se stagnacija poboljsanja
                if child1 < temp_best: temp_best = child1; brstag = 0
                if child2 < temp_best: temp_best = child2; brstag = 0
        P.sort(key=lambda x: (x.korektnost, x.paralelnost))
        self.best = P[0]
        
# Klasa 'paraprofiler' parsira ulazni algoritam te ga dijeli na instrukcijske
# razine koje zatim daje genetskom algoritmu sadrzanom u klasi 'paragen' na
# obradu te na kraju vraca rezultirajuci paralelni algoritam.

class paraprofiler:
    
    def __init__(self, input, *args):
        self.input = self.parser(input)
        self.parametri = args
    
    # Metoda 'parser' parsira ulazni slijedni algoritam.
    
    def parser(self, input):
        input = input.rstrip().split('\n')
        input = [inst.replace(' ', '') for inst in input]
        input = map(lambda x: '"'+x+'"', input)
        for index, inst in enumerate(input):
            if inst[2:5].upper() == 'FOR':
                input[index] = '["'+inst[2:5]+' '+inst[5:-2]+'"'
            elif inst[2:4].upper() == 'IF':
                input[index] = '["'+inst[2:4]+' '+inst[4:-2]+'"'
            elif inst[2:-2].upper() in ('ENDIF', 'ENDFOR'):
                input[index] = ']'
        input = '['+','.join(input)+']'
        return eval(input)
    
    # Metoda 'bernstein_parser' iz za svaku instrukciju (obicnu ili meta) stvara
    # zapis u tablici ulaznih i izlaznih varijabli. Te zapise kasnije koristimo
    # pri evaluaciji dobrote u genetskom algoritmu.
    
    def bernstein_parser(self, input):
        btable = {}
        reAtom = re.compile(r'[a-z_A-Z]+\[i-\d+\]|[a-z_A-Z]+\[\d+\]|[a-z_A-Z]+')
        reIF = re.compile(r'IF\s(([a-z_A-Z]+,)*[a-z_A-Z]+)')
        reFOR = re.compile(r'FOR\s([a-z_A-Z]+).*')
        for index, inst in enumerate(input):
            left, right = set([]), set([])
            if not isinstance(inst, list):
                divide = inst.split('=')
                left = set(reAtom.findall(divide[0]))
                right = set(reAtom.findall(divide[1]))
                btable[index] = (left, right)
            else:
                if inst[0][0:3] == 'FOR':
                    right |= set(reFOR.match(inst[0]).group(1))
                else:
                    right |= set(reIF.match(inst[0]).group(1).split(','))
                result = self.bernstein_parser(inst[1:])
                for key, value in result.items():
                    left |= value[0]
                    right |= value[1]
                btable[index] = (left, right)
        return btable
    
    # Metoda 'result' rekurzivno ide kroz pojedine instrukcijske razine i salje
    # ih u genetski algoritam kako bi dobila konacan raspored instrukcija za
    # svaku od njih. Uz to metoda provodi analizu 'FOR' petlji kako bi znali
    # radi li se o 'FOR' ili 'PARFOR' petlji (kako bi znali mogu li se sve
    # iteracije petlje izvoditi paralelno)
    
    def result(self, *args):
        try:
            input, tabnum, parforFlag = args
        except:
            input, tabnum, parforFlag = self.input, 0, 0
            
        result = ''
        
        def parfor(inst, bernstein_table):
            zapis = bernstein_table[inst]
            regex = re.compile(r'[a-z_A-Z]+\[i-\d+\]')
            varijable = list(zapis[0]|zapis[1])
            for var in varijable:
                if regex.findall(var):
                    return 0
            return 1

        def printinst(inst, tab, paralelno):
            prefix = ''
            parforCheck = 0
            if paralelno == 1 and inst[0][0:3] == 'FOR':
                prefix = 'PAR'
                parforCheck = 1
            return (('%s[%s%s]\n' % ((tab*'    '), prefix, inst[0])) +
                    self.result(inst[1:], tab+1, parforCheck) +
                    ('%s[END%s%s]\n' % ((tab*'    '), prefix, inst[0][0:3])))
        
        # Pokretanje algoritma
        bernstein_table = self.bernstein_parser(input)
        algoritam = paragen(bernstein_table)
        algoritam.parametri(self.parametri)
        algoritam.run()
        najbolja = algoritam.best.zapis
        
        if parforFlag or algoritam.best.korektnost:
            najbolja = [[i] for i in bernstein_table.keys()]
        
        for insts in najbolja:
            if len(insts) == 1:
                inst = input[insts[0]]
                if not isinstance(inst, list):
                    result += ('%s%s\n' % ((tabnum*'    '), inst))
                else:
                    paralelno = parfor(insts[0], bernstein_table)
                    result += printinst(inst, tabnum, paralelno)
            else:
                result += ('%s[PARBEGIN]\n' % (tabnum*'    '))
                for temp_inst in insts:
                    inst = input[temp_inst]
                    if not isinstance(inst, list):
                        result += ('%s%s\n' % ((tabnum*'    '), inst))
                    else:
                        paralelno = parfor(temp_inst, bernstein_table)
                        result += printinst(inst, tabnum, paralelno)
                result += ('%s[PAREND]\n' % (tabnum*'    '))

        return result

if __name__ == "__main__":
    
    # Ovo je dio koda koji omogucava programu da se pokrece u komandnoj
    # liniji ako PyQt nije dostupan.
    try:
        path = raw_input('Upisite putanju do ulazne datoteke:\n')
        datoteka = open(path)
        input = datoteka.read()
    except IOError:
        print 'Navedena datoteka ne postoji!'
        raise SystemExit
    
    try:
        iteracije = int(raw_input('Upisite broj iteracija:\n'))
        stagnacija = int(raw_input('Upisite vrijednost za stagnaciju:\n'))
        populacija = int(raw_input('Upisite velicinu populacije:\n'))
        krizanje = float(raw_input('Upisite vjerojatnost krizanja ([0,1]):\n'))
        if not (0 <= krizanje <= 1):
            print "Vrijednost nije unutar intervala [0,1]!"
            raise SystemExit
        mutacija = float(raw_input('Upisite vjerojatnost mutacije ([0,1]):\n'))
        if not (0 <= mutacija <= 1):
            print "Vrijednost nije unutar intervala [0,1]!"
            raise SystemExit
        maksdubina = int(raw_input('Upisite maksimalnu dubinu stabla:\n'))
    except ValueError:
        print "Jedna ili vise vrijednosti nisu unesene u zadanom formatu!"
        raise SystemExit
    
    algoritam = paraprofiler(
        input,
        iteracije,
        stagnacija,
        populacija,
        krizanje,
        mutacija,
        maksdubina
    )
    print 'Paralelni algoritam:\n'
    print algoritam.result()